<?php

declare(strict_types=1);

namespace Drupal\sm_transport_drupal_sql\Transport;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\DatabaseException;
use Symfony\Component\Messenger\Exception\InvalidArgumentException;
use Symfony\Component\Messenger\Exception\TransportException;

/**
 * Connection to the Drupal database for the SQL transport.
 */
class DrupalSQLConnection {

  /**
   * A set of default options for the connection.
   */
  protected const DEFAULT_OPTIONS = [
    'table_name' => 'messenger_messages',
    'queue_name' => 'default',
    'redeliver_timeout' => 3600,
  ];

  /**
   * DrupalSQLConnection constructor.
   *
   * @param array $configuration
   *   The configuration for the connection.
   * @param \Drupal\Core\Database\Connection $connection
   *   The connection object to the Drupal SQL database.
   */
  public function __construct(
    private readonly Connection $connection,
    protected array $configuration,
  ) {
    $this->configuration = array_replace_recursive(
      static::DEFAULT_OPTIONS,
      $configuration
    );
  }

  /**
   * Returns the configuration for the connection.
   *
   * @return array
   *   The configuration for the connection.
   */
  public function getConfiguration(): array
  {
    return $this->configuration;
  }

  /**
   * Builds the configuration array from the given DSN and options.
   *
   * @param string $dsn
   *   The DSN for the connection.
   * @param array $options
   *   An array of options for the connection.
   *
   * @return array
   *   The configuration array for the connection.
   */
  public static function buildConfiguration(#[\SensitiveParameter] string $dsn, array $options = []): array
  {
    if (false === $params = parse_url($dsn)) {
      throw new InvalidArgumentException('The given Drupal SQL Messenger DSN is invalid.');
    }

    $query = [];
    if (isset($params['query'])) {
      parse_str($params['query'], $query);
    }

    $configuration = [];
    $configuration += $query + $options + static::DEFAULT_OPTIONS;

    return $configuration;
  }

  /**
   * Sends the given Envelope to the Drupal SQL database.
   *
   * This method encodes the Envelope using the serializer and sends the
   * encoded message to the database. If a delay is specified, the message
   * will be available after the delay period.
   * If an exception occurs during the retrieval process, it attempts to
   * ensure the table exists and retries the operation.
   *
   * @param string $body
   *   The body of the message.
   * @param array $headers
   *   The headers of the message.
   * @param int $delay
   *   The delay in milliseconds before the message is available.
   *
   * @return string
   *   The ID of the message in the database.
   *
   * @throws \Exception
   */
  public function send(string $body, array $headers, int $delay = 0): string {
    insert:
    $transaction = $this->connection->startTransaction();
    try {
      $now = new \DateTimeImmutable('UTC');
      $availableAt = $now->modify(sprintf('%+d seconds', $delay / 1000));

      $id = $this->connection->insert($this->configuration['table_name'])
        ->fields([
          'body' => $body,
          'headers' => json_encode($headers),
          'queue_name' => $this->configuration['queue_name'],
          'created_at' => $now->getTimestamp(),
          'available_at' => $availableAt->getTimestamp(),
        ])
        ->execute();
    }
    catch (\Exception $e) {
      $transaction->rollBack();

      // If there was an exception, try to create the table.
      if ($this->ensureTableExists()) {
        // If the exception happened for other reason than the missing table,
        // propagate the exception.
        goto insert;
      }

      throw $e;
    }

    return $id;
  }

  /**
   * Retrieves the next available message from the Drupal SQL database.
   *
   * This method fetches the next message that is available for processing
   * from the database. It marks the message as delivered by updating the
   * `delivered_at` timestamp. If no message is available, it returns NULL.
   * If an exception occurs during the retrieval process, it attempts to
   * ensure the table exists and retries the operation.
   *
   * @return array|null
   *   The decoded message as an associative array, or NULL if no message is
   *   available.
   *
   * @throws \Exception
   */
  public function get(): ?array {
    get:
    $transaction = $this->connection->startTransaction();
    try {
      $now = new \DateTimeImmutable('UTC');
      $redeliverLimit = $now->modify(sprintf('-%d seconds', 3600));

      $query = $this->connection->select($this->configuration['table_name'])
        ->fields($this->configuration['table_name'])
        ->orderBy('available_at')
        ->condition('queue_name', $this->configuration['queue_name'])
        ->condition('available_at', $now->getTimestamp(), '<=');

      $condition = $query->orConditionGroup()
        ->isNull('delivered_at')
        ->condition('delivered_at', $redeliverLimit->getTimestamp(), '<=');
      $query->condition($condition);

      $message = $query
        ->execute()
        ->fetchAssoc();

      if ($message === FALSE) {
        return NULL;
      }

      $doctrineEnvelope = $this->decodeEnvelopeHeaders($message);

      $this->connection->update($this->configuration['table_name'])
        ->fields(['delivered_at' => $now->getTimestamp()])
        ->condition('id', $doctrineEnvelope['id'])
        ->execute();

      return $doctrineEnvelope;
    }
    catch (\Exception $e) {
      $transaction->rollBack();

      // If there was an exception, try to create the table.
      if ($this->ensureTableExists()) {
        // If the exception happened for other reason than the missing table,
        // propagate the exception.
        goto get;
      }

      throw $e;
    }
  }

  /**
   * Acknowledges the message with the given ID by deleting it from the
   * database.
   *
   * This method removes the message with the specified ID from the database,
   * indicating that it has been successfully processed. If an exception occurs
   * during the deletion process, a TransportException is thrown.
   *
   * @param string $id
   *   The ID of the message to acknowledge.
   *
   * @return bool
   *   TRUE if the message was successfully acknowledged, FALSE otherwise.
   *
   * @throws \Symfony\Component\Messenger\Exception\TransportException
   *   Thrown if an error occurs during the acknowledgment process.
   */
  public function ack(string $id): bool {
    try {
      $this
        ->connection
        ->delete($this->configuration['table_name'])
        ->condition('id', $id)
        ->execute();

      return TRUE;
    }
    catch (\Exception $exception) {
      throw new TransportException($exception->getMessage(), 0, $exception);
    }
  }

  /**
   * Rejects the message with the given ID by deleting it from the database.
   *
   * This method removes the message with the specified ID from the database,
   * indicating that it has been rejected and will not be processed. If an
   * exception occurs during the deletion process, a TransportException is
   * thrown.
   *
   * @param string $id
   *   The ID of the message to reject.
   *
   * @return bool
   *   TRUE if the message was successfully rejected, FALSE otherwise.
   *
   * @throws \Symfony\Component\Messenger\Exception\TransportException
   *   Thrown if an error occurs during the rejection process.
   */
  public function reject(string $id): bool {
    try {
      $this
        ->connection
        ->delete($this->configuration['table_name'])
        ->condition('id', $id)
        ->execute();

      return TRUE;
    }
    catch (\Exception $exception) {
      throw new TransportException($exception->getMessage(), 0, $exception);
    }
  }

  /**
   * Ensures that the messenger messages table exists in the database.
   *
   * This method attempts to create the table defined by the schema if it does
   * not already exist. If the table creation fails due to an exception other
   * than the table already existing, it returns FALSE. Otherwise, it returns
   * TRUE.
   *
   * @return bool
   *   TRUE if the table exists or was successfully created, FALSE otherwise.
   */
  protected function ensureTableExists(): bool {
    try {
      $database_schema = $this->connection->schema();
      $schema_definition = $this->schemaDefinition();
      $database_schema->createTable(
        $this->configuration['table_name'],
        $schema_definition
      );
    }
      // If another process has already created the messenger table, attempting
      // to recreate it will throw an exception. In this case just catch the
      // exception and do nothing.
    catch (DatabaseException $e) {
    }
    catch (\Exception $e) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Defines the schema for the messenger_messages table.
   *
   * @return array
   *   An associative array defining the schema for the messenger_messages
   *   table.
   *
   * @internal
   */
  public function schemaDefinition(): array {
    return [
      'description' => 'Store Messenger messages.',
      'fields' => [
        'id' => [
          'type' => 'serial',
          'size' => 'big',
          'not null' => TRUE,
        ],
        'body' => [
          'type' => 'text',
          'size' => 'big',
          'not null' => TRUE,
        ],
        'headers' => [
          'type' => 'text',
          'size' => 'big',
          'not null' => TRUE,
        ],
        'queue_name' => [
          'type' => 'varchar',
          'length' => 190,
          'not null' => TRUE,
        ],
        'created_at' => [
          'type' => 'int',
          'not null' => TRUE,
          'size' => 'big',
        ],
        'available_at' => [
          'type' => 'int',
          'not null' => TRUE,
          'size' => 'big',
        ],
        'delivered_at' => [
          'type' => 'int',
          'not null' => FALSE,
        ],
      ],
      'primary key' => ['id'],
      'indexes' => [
        'delivered_at' => ['delivered_at'],
        'available_at' => ['available_at'],
        'queue_name' => ['queue_name'],
      ],
    ];
  }

  /**
   * Decodes the headers of the given message envelope.
   *
   * This method takes an associative array representing a message envelope
   * and decodes the JSON-encoded headers into an associative array.
   *
   * @param array $doctrineEnvelope
   *   The message envelope with JSON-encoded headers.
   *
   * @return array
   *   The message envelope with decoded headers.
   */
  private function decodeEnvelopeHeaders(array $doctrineEnvelope): array {
    $doctrineEnvelope['headers'] = json_decode(
      $doctrineEnvelope['headers'],
      TRUE
    );

    return $doctrineEnvelope;
  }

}
