<?php

declare(strict_types=1);

namespace Drupal\sm_transport_drupal_sql\Transport;

use Symfony\Component\Messenger\Envelope;
use Symfony\Component\Messenger\Exception\LogicException;
use Symfony\Component\Messenger\Exception\MessageDecodingFailedException;
use Symfony\Component\Messenger\Exception\TransportException;
use Symfony\Component\Messenger\Stamp\TransportMessageIdStamp;
use Symfony\Component\Messenger\Transport\Receiver\ReceiverInterface;
use Symfony\Component\Messenger\Transport\Serialization\PhpSerializer;
use Symfony\Component\Messenger\Transport\Serialization\SerializerInterface;

/**
 * A receiver for the Drupal SQL transport.
 */
class DrupalSQLReceiver implements ReceiverInterface {

  /**
   * A serializer to decode the message data.
   *
   * @var \Symfony\Component\Messenger\Transport\Serialization\SerializerInterface
   */
  private SerializerInterface $serializer;

  /**
   * DrupalSQLReceiver constructor.
   *
   * @param \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLConnection $connection
   *   The DrupalSQLConnection object.
   * @param \Symfony\Component\Messenger\Transport\Serialization\SerializerInterface|null $serializer
   *   The serializer to use.
   */
  public function __construct(
    private readonly DrupalSQLConnection $connection,
    ?SerializerInterface $serializer = NULL,
  ) {
    $this->serializer = $serializer ?? new PhpSerializer();
  }

  /**
   * {@inheritdoc}
   */
  public function get(): iterable {
    try {
      $doctrineEnvelope = $this->connection->get();
    }
    catch (\Exception $exception) {
      throw new TransportException($exception->getMessage(), 0, $exception);
    }

    if (NULL === $doctrineEnvelope) {
      return [];
    }

    return [$this->createEnvelopeFromData($doctrineEnvelope)];
  }

  /**
   * {@inheritdoc}
   */
  public function ack(Envelope $envelope): void {
    $this->connection->ack(
      $this->findDrupalSQLReceivedStamp($envelope)->getId()
    );
  }

  /**
   * {@inheritdoc}
   */
  public function reject(Envelope $envelope): void {
    $this->connection->ack(
      $this->findDrupalSQLReceivedStamp($envelope)->getId()
    );
  }

  /**
   * Creates an Envelope object from the provided data array.
   *
   * This method decodes the message data using the serializer and constructs
   * an Envelope object. If the message decoding fails, the message is
   * rejected.
   * The Envelope is then returned with the appropriate stamps.
   *
   * @param array $data
   *   The data array containing the message body and headers.
   *
   * @return \Symfony\Component\Messenger\Envelope
   *   The constructed Envelope object.
   *
   * @throws \Symfony\Component\Messenger\Exception\MessageDecodingFailedException
   *   Thrown if the message decoding fails.
   */
  private function createEnvelopeFromData(array $data): Envelope {
    try {
      $envelope = $this->serializer->decode([
        'body' => $data['body'],
        'headers' => $data['headers'],
      ]);
    }
    catch (MessageDecodingFailedException $exception) {
      $this->connection->reject($data['id']);

      throw $exception;
    }

    return $envelope->with(
      new DrupalSQLReceivedStamp($data['id']),
      new TransportMessageIdStamp($data['id'])
    );
  }

  /**
   * Retrieves the DrupalSQLReceivedStamp from the provided Envelope.
   *
   * @param \Symfony\Component\Messenger\Envelope $envelope
   *   The Envelope object to retrieve the DrupalSQLReceivedStamp from.
   *
   * @return \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLReceivedStamp
   *   The DrupalSQLReceivedStamp object.
   *
   * @throws \Symfony\Component\Messenger\Exception\LogicException
   *   Thrown if no DrupalSQLReceivedStamp is found on the Envelope.
   */
  private function findDrupalSQLReceivedStamp(
    Envelope $envelope,
  ): DrupalSQLReceivedStamp {
    /** @var DrupalSQLReceivedStamp|null $drupalSQLReceivedStamp */
    $drupalSQLReceivedStamp = $envelope->last(DrupalSQLReceivedStamp::class);

    if (NULL === $drupalSQLReceivedStamp) {
      throw new LogicException(
        'No DrupalSQLReceivedStamp found on the Envelope.'
      );
    }

    return $drupalSQLReceivedStamp;
  }

}
