<?php

declare(strict_types=1);

namespace Drupal\sm_transport_drupal_sql\Transport;

use Drupal\Core\Database\Connection;
use Symfony\Component\Messenger\Envelope;
use Symfony\Component\Messenger\Transport\Serialization\SerializerInterface;
use Symfony\Component\Messenger\Transport\TransportInterface;

/**
 * A transport implementation for handling messages via Drupal SQL.
 *
 * This class provides methods to send, acknowledge, and reject messages
 * using a Drupal SQL database as the transport mechanism. It utilizes
 * the `DrupalSQLSender` and `DrupalSQLReceiver` classes to perform
 * the actual sending and receiving of messages.
 *
 */
class DrupalSQLTransport implements TransportInterface {

  /**
   * An internal connection object to the Drupal SQL database.
   *
   * @var \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLConnection
   */
  private DrupalSQLConnection $connection;

  /**
   * The receiver object for this transport.
   *
   * @var \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLReceiver
   */
  private DrupalSQLReceiver $receiver;

  /**
   * The sender object for this transport.
   *
   * @var \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLSender
   */
  private DrupalSQLSender $sender;

  /**
   * DrupalSQLTransport constructor.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The connection object to the Drupal SQL database.
   * @param \Symfony\Component\Messenger\Transport\Serialization\SerializerInterface $serializer
   *   The serializer to use for encoding and decoding messages.
   */
  public function __construct(
    Connection $connection,
    private readonly SerializerInterface $serializer,
    array $configuration,
  ) {
    $this->connection = new DrupalSQLConnection($connection, $configuration);
  }

  /**
   * {@inheritdoc}
   */
  public function get(): iterable {
    return $this->getReceiver()->get();
  }

  /**
   * {@inheritdoc}
   */
  public function ack(Envelope $envelope): void {
    $this->getReceiver()->ack($envelope);
  }

  /**
   * {@inheritdoc}
   */
  public function reject(Envelope $envelope): void {
    $this->getReceiver()->reject($envelope);
  }

  /**
   * {@inheritdoc}
   */
  public function send(Envelope $envelope): Envelope {
    return $this->getSender()->send($envelope);
  }

  /**
   * Return the receiver object for this transport.
   *
   * @return \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLReceiver
   *   The receiver object.
   */
  private function getReceiver(): DrupalSQLReceiver {
    return $this->receiver ??= new DrupalSQLReceiver(
      $this->connection,
      $this->serializer
    );
  }

  /**
   * Return the sender object for this transport.
   *
   * @return \Drupal\sm_transport_drupal_sql\Transport\DrupalSQLSender
   *   The sender object.
   */
  private function getSender(): DrupalSQLSender {
    return $this->sender ??= new DrupalSQLSender(
      $this->connection,
      $this->serializer
    );
  }

}
